(*==================*)
(*                  *)
(*=== ILL in Coq ===*)
(*                  *)
(*==================*)

(* v.0.5.6 Olivier Laurent *)
 
(* Thanks to Damien Pous and Lionel Rieg *)



(*== Preliminary tools ==*)

Ltac no_hyps :=
  (* test if the linear context is empty *)
  lazymatch goal with
  | H : ?A |- _ => lazymatch A with
                   | ?t => lazymatch type of t with
                           | Set  => revert H ; no_hyps ; intro H
                           | Type => revert H ; no_hyps ; intro H
                           end
                   | _ => fail "The context is not empty"
                   end
  | _           => idtac
  end.

Ltac protect t :=
  (* apply t after protecting hypotheses of sort Set and Type by reverting them *)
  match goal with
  | H : ?A |- _ => lazymatch type of A with
                   | Set  => revert H ; (protect t) ; intro H at top
                   | Type => revert H ; (protect t) ; intro H at top
                   end
  | _           => t
  end.
(* known bug :
   If t = clean - G, with an element Ha not in G depending on some a : Prop,
   then a is erased.
   Goal forall a b : Prop, a -> b -> b.
   intros a b Ha Hb.
   protect ltac:(clear - Hb).
*)

Ltac lin_clear H :=
  (* clear hypothesis H if it is of sort Set or Type *)
  let A := type of H in
  lazymatch type of A with
  | Set  => clear H
  | Type => clear H
  | _    => fail "Hypothesis is not of sort Set or Type"
  end.

Ltac lin_clear_props :=
  (* remove useless hypotheses of type Prop *)
  match goal with
  | H : Prop |- _ => (clear H ; lin_clear_props)
                     || (revert H ; lin_clear_props ; intro H)
  | _             => idtac
  end.

Ltac lin_clear_types :=
  (* remove useless hypotheses of sort Type *)
  match goal with
  | H : ?A |- _ => lazymatch type of A with
                   | Type => (clear H ; lin_clear_types)
                            || (revert H ; lin_clear_types ; intro H)
                   | _    => revert H ; lin_clear_types ; intro H
                   end
  | _           => idtac
  end.

Ltac lin_clear_sets :=
  (* remove useless hypotheses of sort Set *)
  match goal with
  | H : ?A |- _ => lazymatch type of A with
                   | Set => (clear H ; lin_clear_sets)
                            || (revert H ; lin_clear_sets ; intro H)
                   | _   => revert H ; lin_clear_sets ; intro H
                   end
  | _           => idtac
  end.

Ltac lin_clear_all :=
  (* remove useless hypotheses of sort Set or Type *)
  match goal with
  | H : ?A |- _ => lazymatch type of A with
                   | Set  => (clear H ; lin_clear_all)
                            || (revert H ; lin_clear_all ; intro H)
                   | Type => (clear H ; lin_clear_all)
                            || (revert H ; lin_clear_all ; intro H)
                   | _    => revert H ; lin_clear_all ; intro H
                   end
  | _           => idtac
  end.


(*== Splitting tacticals ==*)

Ltac lin_bin_split_general t b1 cl1 e1 b2 cl2 e2 :=
  (* apply binary tactic t followed in each branch by
     b_i then cl_i with protection and then e_i *)
  t ; [ b1 ; protect cl1 ; e1 | b2 ; protect cl2 ; e2 ].

Ltac lin_bin_split t cl1 cl2 :=
  (* apply binary tactic t followed by cl1 and cl2 with protection *)
  lin_bin_split_general t idtac cl1 idtac idtac cl2 idtac.

Ltac lin_ter_split t cl1 cl2 cl3 :=
  (* apply ternary tactic t followed by cl1, cl2 and cl3 with protection *)
  t ; [ protect cl1 | protect cl2 | protect cl3 ].

Ltac lin_qua_split t cl1 cl2 cl3 cl4 :=
  (* apply 4-ary tactic t followed by cl1, cl2, cl3 and cl4 with protection *)
  t ; [ protect cl1 | protect cl2 | protect cl3 | protect cl4 ].


(*== Axiom and Cut ==*)

Ltac lin_assumption :=
  (* axiom rule (requires singleton context) *)
  lazymatch goal with
  | H : ?A |- ?A => revert H ; no_hyps ; intro H ;
                    exact H
  | H : _  |- _  => fail "Assumptions are different from goal"
  | _            => fail "No hypothesis found"
  end.

Tactic Notation
  (* cut rule *)
  "lin_cut" constr(c) "as" simple_intropattern(h) "with" hyp_list(G) :=
    lin_bin_split_general ltac:(cut c)
       ltac:(idtac) ltac:(clear - G) ltac:(intro h)
       ltac:(idtac) ltac:(clear G) ltac:(idtac).


(*== Multiplicatives ==*)

Notation "A -o B" := (A -> B) (right associativity, at level 60) : mll_scope.
Notation "1" := True : mll_scope.
Notation "A  (x)  B" := (A /\ B) (at level 40) : mll_scope.
Notation "A o-o B" := (A <-> B) (at level 65) : mll_scope.

Ltac lin_one :=
  (* right one rule *)
  no_hyps ; split ; fail "Context is not empty".

Ltac lin_left_one H :=
  (* left one rule *)
  lazymatch type of H with
  | True => destruct H
  | _    => fail "Hypothesis is not True"
  end.

Tactic Notation "lin_split" "with" hyp_list(G) :=
  (* tensor right rule *)
  lin_bin_split ltac:(split) ltac:(clear - G) ltac:(clear G).

Ltac lin_left_conj t :=
  (* tensor left rule *)
  (let l := fresh t in
   let r := fresh t in
     destruct t as [l r])
  ||
  (let l := fresh in
   let r := fresh in
     destruct t as [l r]).

Ltac lem_apply0 L :=
  (* 0-ary use of a lemma *)
  no_hyps ; apply L ; fail "Wrong number of tactics".

Ltac lem_apply1 L :=
  (* 1-ary use of a lemma *)
  apply L ; [idtac].

Tactic Notation
  (* 2-ary use of a lemma *)
  "lem_apply2" constr(L) "with" hyp_list(G) :=
    lin_bin_split ltac:(apply L) ltac:(clear - G) ltac:(clear G).

Tactic Notation
  (* 3-ary use of a lemma *)
  "lem_apply3" constr(L) "with" hyp_list(G1) "::" hyp_list(G2) :=
    lin_ter_split ltac:(apply L)
      ltac:(clear - G1) ltac:(clear - G2) ltac:(clear G1 ; clear G2).

Tactic Notation
  (* 4-ary use of a lemma *)
  "lem_apply4" constr(L) "with" hyp_list(G1) "::" hyp_list(G2) "::" hyp_list(G3) :=
    lin_qua_split ltac:(apply L)
      ltac:(clear - G1) ltac:(clear - G2) ltac:(clear - G3)
      ltac:(clear G1 ; clear G2 ; clear G3).

Tactic Notation
  (* 1-ary use of a lemma on an hypothesis *)
  "lem_hyp_apply1" constr(L) "in" constr(H) :=
    simple apply L in H ; [idtac].

Tactic Notation
  (* 2-ary use of lemma on an hypothesis *)
  "lem_hyp_apply2" constr(L) "in" constr(H) "with" hyp_list(G) :=
    lin_bin_split ltac:(simple apply L in H) ltac:(clear G) ltac:(clear - G).

Tactic Notation
  (* 0-ary use of an hypothesis
  (same as lin_assumption but allows for automatic instantiation of variables) *)
  "lin_apply0" hyp(H) :=
    revert H ; no_hyps ; intro H ; simple apply H ;
    fail "Wrong number of tactics".

Tactic Notation 
  (* 1-ary use of an hypothesis *)
  "lin_apply1" hyp(H) :=
    simple apply H ; clear H ; [idtac].

Tactic Notation
  (* 2-ary use of an hypothesis *)
  "lin_apply2" hyp(H) "with" hyp_list(G) :=
    lin_bin_split ltac:(simple apply H) ltac:(clear - G) ltac:(clear G ; clear H).

Tactic Notation
  (* 3-ary use of an hypothesis *)
  "lin_apply3" hyp(H) "with" hyp_list(G1) "::" hyp_list(G2) :=
    lin_ter_split ltac:(simple apply H)
      ltac:(clear - G1) ltac:(clear - G2) ltac:(clear G1 ; clear G2 ; clear H).

Tactic Notation
  (* 4-ary use of an hypothesis *)
  "lin_apply4" hyp(H) "with" hyp_list(G1) "::" hyp_list(G2) "::" hyp_list(G3) :=
    lin_qua_split ltac:(simple apply H)
      ltac:(clear - G1) ltac:(clear - G2) ltac:(clear - G3)
      ltac:(clear G1 ; clear G2 ; clear G3 ; clear H).

Tactic Notation
  (* 1-ary use of an hypothesis in another hypothesis *)
  "lin_hyp_apply1" hyp(L) "in" constr(H) :=
    simple apply L in H ; clear L ; [idtac].

Tactic Notation
  (* 2-ary use of an hypothesis in another hypothesis *)
  "lin_hyp_apply2" hyp(L) "in" constr(H) "with" hyp_list(G) :=
    lin_bin_split ltac:(simple apply L in H) ltac:(clear L ; clear G) ltac:(clear - G).

Tactic Notation
  (* left linear implication rule *)
  "lin_left_imp" hyp(H) "with" hyp_list(G) :=
    lin_bin_split_general ltac:(lapply H)
       ltac:(idtac) ltac:(clear H ; clear - G) ltac:(intro H)
       ltac:(idtac) ltac:(clear H ; clear G) ltac:(idtac).


(*== Additives ==*)

Inductive ltop : Prop :=
  (* T connective *)
| ltop_c : ltop.
Notation "[T]" := ltop.

Ltac lin_top :=
  (* right top rule *)
  lazymatch goal with
  | |- ltop => split
  | _       => fail "Conclusion is not [T]"
  end.

Inductive lwith (A B : Prop) : Prop :=
  (* & connective *)
| lwith_c : A -> B -> lwith A B.
Notation "A  [&]  B" := (lwith A B) (at level 45).

Ltac lin_with :=
  (* right with rule *)
  lazymatch goal with
  | |- lwith _ _ => split
  | _            => fail "Conclusion is not a [&]"
  end.

Ltac lin_left_with_1 H :=
  (* left with rule 1 *)
  lazymatch type of H with
  | lwith _ _ => let H1 := fresh H in
                 let H2 := fresh H in
                 destruct H as [H1 H2] ;
                 clear H2 ;
                 rename H1 into H
  | _         => fail "Hypothesis is not a [&]"
  end.

Ltac lin_left_with_2 H :=
  (* left with rule 2 *)
  lazymatch type of H with
  | lwith _ _ => let H1 := fresh H in
                 let H2 := fresh H in
                 destruct H as [H1 H2] ;
                 clear H1 ;
                 rename H2 into H
  | _         => fail "Hypothesis is not a [&]"
  end.

Inductive lzero : Prop := . (* 0 connective *)
Notation "[0]" := lzero.

Ltac lin_left_zero H :=
  (* left zero rule *)
  lazymatch type of H with
  | lzero => destruct H
  | _     => fail "Hypothesis is not a [0]"
  end.

Inductive lplus (A B : Prop) : Prop :=
  (* plus connective *)
| lplus_c_1 : A -> lplus A B
| lplus_c_2 : B -> lplus A B.
Notation "A  [+]  B" := (lplus A B) (at level 45).

Ltac lin_plus_1 :=
  (* right plus rule 1 *)
  lazymatch goal with
  | |- lplus _ _ => constructor 1
  | _            => fail "Hypothesis is not a [+]"
  end.

Ltac lin_plus_2 :=
  (* right plus rule 2 *)
  lazymatch goal with
  | |- lplus _ _ => constructor 2
  | _            => fail "Hypothesis is not a [+]"
  end.

Ltac lin_left_plus H :=
  (* left plus rule *)
  lazymatch type of H with
  | lplus _ _ => destruct H as [H | H]
  | _         => fail "Hypothesis is not a [+]"
  end.

Notation "'T'" := ltop : all_scope.
Notation "A  (&)  B" := (lwith A B) (at level 45) : all_scope.
Notation "0" := lzero : all_scope.
Notation "A  (+)  B" := (lplus A B) (at level 45) : all_scope.

Definition addiff (A B : Prop) := lwith (A -> B) (B -> A).
  (* linear equivalence based on additive conjunction *)
Notation "A o-&-o B" := (addiff A B) (at level 66).

Ltac lin_split_equiv :=
  (* split both kinds of linear equivalences *)
  match goal with
  | |- addiff _ _ => unfold addiff ; lin_with
  | |- iff _ _ => lin_split with
  end.


(*== Exponentials ==*)

Parameter OC : Prop -> Prop.
   (* ! connective *)
Notation "! A" := (OC A) (at level 31).

(*= Weakening *)

Ltac lin_weak H :=
  (* left weakening rule *)
  lazymatch type of H with
  | OC _ => clear H
  end
  || fail "Hypothesis does not start with !".

Ltac lin_clear_oc :=
  (* remove all hypotheses of the shape !_ *)
  lazymatch goal with
  | H : OC _ |- _ => lin_weak H ; lin_clear_oc
  | _             => idtac
  end.

Ltac lin_assumption_oc :=
  (* axiom rule with !_ context *)
  lazymatch goal with
  | H : ?A |- ?A => revert H ;
                    lin_clear_oc ;
                    no_hyps ;
                    intro H ;
                    exact H
  | H : _  |- _  => fail "Assumptions are different from goal"
  | _            => fail "No hypothesis found"
  end.

Ltac lin_one_oc :=
  (* right one rule with !_ context *)
  lin_clear_oc ; lin_one.

(*= Contraction =*)

Ltac lin_contr H :=
  (* left contraction rule *)
  let H0 := fresh H in 
    lazymatch type of H with
    | OC _ => assert(H0 := H)
    end
    || fail "Hypothesis does not start with !".


(*== Exponentials for LL ==*)

(*= Dereliction =*)

Parameter oc_der : forall X : Prop, OC X -> X.
  (* dereliction principle, to be mainly used in lin_der and lin_right_der tactics *)

Ltac lin_der H := lem_hyp_apply1 oc_der in H.
  (* left dereliction rule *)
Ltac lin_right_der := lem_apply1 oc_der.
  (* elimination rule for ! on the right *)

Ltac lin_assumption_der :=
  (* axiom rule with !_ context up to dereliction rule *)
  lazymatch goal with
  | H : OC ?A |- ?A => revert H ;
                       lin_clear_oc ;
                       no_hyps ;
                       intro H ;
                       lin_der H ;
                       exact H
  | H : _  |- _     => fail "Assumptions do not match goal"
  | _               => fail "No hypothesis found"
  end.

(*= Usual promotion =*)

Parameter addoc_FORBIDDEN_IN_PROOFS : forall a : Prop, a -> OC a.
  (* wrong principle to be used in promotion tactics only *)

Ltac are_oc_hyps := (* test if all hypotheses of sort Prop start with ! *)
  lazymatch goal with
  | H : ?A |- _ => lazymatch A with
                   | OC _ => revert H ; are_oc_hyps ; intro H
                   | ?t   => lazymatch type of t with
                             | Set  => revert H ; are_oc_hyps ; intro H
                             | Type => revert H ; are_oc_hyps ; intro H
                             end
                   | _    => fail "Hypothesis does not start with !"
                   end
  | _           => idtac
  end.

Ltac remoc_goal k :=
  (* removes external ! in goal using k : forall a : Prop, a -> !a *)
  lazymatch goal with
  | |- OC _ => simple apply k
  | _       => fail "Conclusion does not start with !"
  end.

Ltac lin_prom := (* usual promotion rule *)
  are_oc_hyps ;
  remoc_goal addoc_FORBIDDEN_IN_PROOFS.

(*= Multi-functorial Promotion for LL =*)

Ltac oc_hyps k :=
  (* act on all hypotheses of sort Prop which must start with ! 
     by k : forall a : Prop, !a -> _ *)
  lazymatch goal with
  | H : ?A |- _ => lazymatch A with
                   | OC _ => simple apply k in H ;
                             revert H ;
                             oc_hyps k ;
                             intro H
                   | ?t   => lazymatch type of t with
                             | Set  => revert H ; oc_hyps k ; intro H
                             | Type => revert H ; oc_hyps k ; intro H
                             end
                   | _    => fail "Hypothesis does not start with !"
                   end
  | _           => idtac
  end.

Ltac lin_mfprom := (* multi-functorial promotion rule for LL *)
  lin_prom ;
  oc_hyps oc_der.

(*= Digging =*)

Lemma oc_dig : forall a : Prop, OC a -> OC (OC a).
  (* digging principle *)
Proof.
intros a o.
lin_prom.
lin_assumption.
Qed.

Ltac lin_dig H := lem_hyp_apply1 oc_dig in H.
  (* left digging rule *)
Ltac lin_right_dig := lem_apply1 oc_dig.
  (* elimination rule !! from ! on the right *)


(*== ELL / LLL ==*)

(*= Promotion rules for ! =*)

Parameter oc_fprom_FORBIDDEN_IN_PROOFS : forall a b : Prop, (a -> b) -> OC a -> OC b.
  (* Invalid promotion principle, to be used in an empty context only *)
Parameter oc_mon_m : forall a b : Prop, OC a -> OC b -> OC (a /\ b).
Parameter oc_mon_n : True -> OC True.
  (* Monoidality principles *)

Ltac lin_fprom_oc :=
  (* functorial promotion *)
  lazymatch goal with
  | H : OC _ |- _ => revert H ;
                     no_hyps ;
                     apply oc_fprom_FORBIDDEN_IN_PROOFS ;
                     intro H
  end.

Ltac lin_ufprom_oc :=
  (* functorial promotion allowing empty exponential context *)
  lazymatch goal with
  | H : OC _ |- _ => revert H ;
                     no_hyps ;
                     apply oc_fprom_FORBIDDEN_IN_PROOFS ;
                     intro H
  | _ => let H := fresh in
         assert (!True) as H ; [ lem_apply1 oc_mon_n ; lin_one | idtac] ;
         lin_ufprom_oc ;
         lin_left_one H
  end.

Ltac lin_mfprom_oc :=
  (* multifunctorial promotion *)
  lazymatch goal with
  | H1 : OC _, H2 : OC _ |- _ => simple apply (oc_mon_m _ _ H1) in H2 ; clear H1 ;
                                 lin_mfprom_oc ;
                                 destruct H2 as [H1 H2]
  | H : OC _ |- _             => revert H ;
                                 no_hyps ;
                                 apply oc_fprom_FORBIDDEN_IN_PROOFS ;
                                 intro H
  | _                         => let H := fresh in
                                 assert (!True) as H ;
                                   [ lem_apply1 oc_mon_n ; lin_one | idtac] ;
                                 lin_mfprom_oc ;
                                 lin_left_one H
  end.

(*= Promotion rules for § =*)

Parameter PG : Prop -> Prop.
  (* § connective *)
Notation "§ A" := (PG A) (at level 31).

Parameter pg_fprom_FORBIDDEN_IN_PROOFS : forall a b : Prop, (a -> b) -> PG a -> PG b.
  (* Invalid promotion principle, to be used in an empty context only *)
Parameter pg_mon_m : forall a b : Prop, PG a -> PG b -> PG (a /\ b).
Parameter pg_mon_n : True -> PG True.
  (* Monoidality principles *)
Parameter pg_der : forall a : Prop, OC a -> PG a.
  (* Weak dereliction principle *)

Ltac are_pg_hyps := (* test if all hypotheses of sort Prop start with § *)
  lazymatch goal with
  | H : ?A |- _ => lazymatch A with
                   | PG _ => revert H ; are_pg_hyps ; intro H
                   | ?t   => lazymatch type of t with
                             | Set  => revert H ; are_pg_hyps ; intro H
                             | Type => revert H ; are_pg_hyps ; intro H
                             end
                   | _    => fail "Hypothesis does not start with PG"
                   end
  | _           => idtac
  end.

Ltac lin_mfprom_pg :=
  (* multifunctorial promotion *)
  lazymatch goal with
  | H : OC _ |- _             => simple apply pg_der in H ; lin_mfprom_pg
  | H1 : PG _, H2 : PG _ |- _ => simple apply (pg_mon_m _ _ H1) in H2 ; clear H1 ;
                                 lin_mfprom_pg ;
                                 destruct H2 as [H1 H2]
  | H : PG _ |- _             => revert H ;
                                 no_hyps ;
                                 apply pg_fprom_FORBIDDEN_IN_PROOFS ;
                                 intro H
  | _                         => let H := fresh in
                                 assert (§True) as H ;
                                   [ lem_apply1 pg_mon_n ; lin_one | idtac] ;
                                 lin_mfprom_pg ;
                                 lin_left_one H
  end.

Ltac lin_der_pg H := lem_hyp_apply1 pg_der in H.
  (* left dereliction rule for § *)
Ltac lin_right_der_pg := lem_apply1 pg_der.
  (* elimination rule for !§ on the right *)


(*= Connective # =*)

Parameter SH : Prop -> Prop.
  (* # connective *)
Notation "# A" := (SH A) (at level 31).

Parameter sh_dig : forall a : Prop, OC a -> OC (SH a).
  (* digging principle with # *)
Parameter sh_der : forall a : Prop, SH a -> a.
  (* dereliction principle for # *)

Ltac multi_digging_sh :=
  (* application of the digging principle on all hypotheses *)
  lazymatch goal with
  | H : OC _ |- _  => simple apply sh_dig in H ;
                      revert H ;
                      multi_digging_sh ;
                      intro H
  | _              => idtac
  end.

Ltac lin_mfprom_sh :=
  (* promotion rule for ! with # *)
  multi_digging_sh ;
  lin_mfprom_oc
.

Ltac lin_der_sh H := lem_hyp_apply1 sh_der in H.
  (* left dereliction rule for # *)

Ltac lin_contr_sh H :=
  (* left contraction rule for # *)
  let H0 := fresh H in 
    lazymatch type of H with
    | SH _ => assert(H0 := H)
    end
    || fail "Hypothesis does not start with #".

Ltac lin_weak_sh H :=
  (* left weakening rule for # *)
  lazymatch type of H with
  | SH _ => clear H
  end
  || fail "Hypothesis does not start with #".

Ltac lin_clear_sh :=
  (* remove all hypotheses of the shape #_ *)
  lazymatch goal with
  | H : SH _ |- _ => lin_weak_sh H ; lin_clear_sh
  | _             => idtac
  end.

Ltac lin_assumption_sh :=
  (* axiom rule with #_ and !_ context *)
  lazymatch goal with
  | H : ?A |- ?A => revert H ;
                    lin_clear_oc ;
                    lin_clear_sh ;
                    no_hyps ;
                    intro H ;
                    exact H
  | H : _  |- _  => fail "Assumptions are different from goal"
  | _            => fail "No hypothesis found"
  end.

Ltac lin_one_sh :=
  (* right one rule with #_ and !_ context *)
  lin_clear_oc ; lin_clear_sh ; lin_one.

Ltac lin_assumption_der_sh :=
  (* axiom rule with #_ and !_ context up to dereliction_sh rule *)
  lazymatch goal with
  | H : SH ?A |- ?A => revert H ;
                       lin_clear_sh ;
                       no_hyps ;
                       intro H ;
                       lin_der_sh H ;
                       exact H
  | H : _  |- _  => fail "Assumptions do not match goal"
  | _            => fail "No hypothesis found"
  end.


(*== SLL ==*)

Parameter ocsh_der : forall a : Prop, OC a -> SH a.
  (* Coercion from ! to # to start a multiplexing rule *)
Ltac lin_der_sll H := lem_hyp_apply1 ocsh_der in H.
  (* left coercion rule from ! to # *)

Ltac lin_multiplex_sh n H :=
  (* n ary left multiplexing rule *)
  lazymatch n with
  | 0    => lin_weak_sh H
  | 1    => lin_der_sh H
  | S ?k => let H0 := fresh H in 
              (lazymatch type of H with
               | SH _ => assert(H0 := H)
               end
               || fail "Hypothesis does not start with #") ;
              lin_der_sh H0 ;
              lin_multiplex_sh k H
  end.
Ltac lin_multiplex n H :=
  (* multiplexing rule of rank n *)
  lin_der_sll H ; lin_multiplex_sh n H.

Fixpoint Tensit (n:nat) (A:Prop) :=
  (* iterated tensor product *)
  match n with
  | 0 => True
  | 1 => A
  | S k => A /\ Tensit k A
  end.
Parameter multiplex : forall n : nat, forall a : Prop, !a -> Tensit n a.
  (* axiomatic definition of the multiplexing principle *)


(*== Iterated modalities ==*)

Require Arith.

(*= Iterated ! =*)

(* with implicit parameter
Fixpoint OCit {n:nat} (A:Prop) :=
  match n with
  | 0 => A
  | S k => OC (@OCit k A)
  end.
*)

Fixpoint OCit (n:nat) (A:Prop) :=
  match n with
  | 0 => A
  | S k => OC (OCit k A)
  end.
Notation "!^( n ) A" := (OCit n A) (at level 32).

Fact ocit_one : forall a : Prop, !a = !^(1)a.
Proof.
intro a.
simpl.
reflexivity.
Qed.

Fact ocit_add : forall n m : nat, forall a : Prop, (!^(n)!^(m)a) = (!^(n+m)a).
Proof.
intros n m a.
induction n ;
simpl.
reflexivity.
rewrite IHn.
reflexivity.
Qed.

Fact ocit_S : forall n : nat, forall a : Prop, !^(n)!a = !^(n+1)a.
Proof.
induction n ;
simpl.
reflexivity.
intro a.
rewrite IHn.
reflexivity.
Qed.

Fact ococit_ocitoc : forall n : nat, forall a : Prop, (!!^(n)a) = !^(n)!a.
Proof.
intros n a.
simpl.
rewrite ocit_one.
rewrite ocit_add.
rewrite Coq.Arith.Plus.plus_comm.
rewrite ocit_S.
reflexivity.
Qed.


(*= Iterated § =*)

Fixpoint PGit (n:nat) (A:Prop) :=
  match n with
  | 0 => A
  | S k => PG (PGit k A)
  end.
Notation "§^( n ) A" := (PGit n A) (at level 32).

Fact pgit_one : forall a : Prop, §a = §^(1)a.
Proof.
intro a.
simpl.
reflexivity.
Qed.

Fact pgit_add : forall n m : nat, forall a : Prop, (§^(n)§^(m)a) = (§^(n+m)a).
Proof.
intros n m a.
induction n ;
simpl.
reflexivity.
rewrite IHn.
reflexivity.
Qed.

Fact pgit_S : forall n : nat, forall a : Prop, §^(n)§a = §^(n+1)a.
Proof.
induction n ;
simpl.
reflexivity.
intro a.
rewrite IHn.
reflexivity.
Qed.

Fact pgpgit_pgitpg : forall n : nat, forall a : Prop, (§§^(n)a) = §^(n)§a.
Proof.
intros n a.
simpl.
rewrite pgit_one.
rewrite pgit_add.
rewrite Coq.Arith.Plus.plus_comm.
rewrite pgit_S.
reflexivity.
Qed.


(*= Iterated ! and § =*)

Fact pgit_der : forall n : nat, forall a : Prop, !^(n)a -> §^(n)a.
Proof.
intros n a.
induction n. (* meta induction *)
simpl.
intro Ha ;
lin_assumption.
simpl.
intro Ho.
lin_der_pg Ho.
revert Ho.
apply pg_fprom_FORBIDDEN_IN_PROOFS.
assumption.
Qed.

(*
Ltac pgit_der_tac := (* construction of the proofs of pgit_der: !^(k)a -> §^(k)a
                         for a given k *)
  simpl ;
  intro O ;
  (repeat lin_mfprom_pg) ; 
  lin_assumption.
*)

(*= Iterated promotion for LL =*)

Parameter addocit_FORBIDDEN_IN_PROOFS : forall n : nat, forall a : Prop, a -> OCit n a.
  (* wrong principle to be used in promotion tactics only *)

Ltac remocit_goal k :=
  (* removes external !^(n) in goal using k : forall a : Prop, a -> !^(n)a *)
  lazymatch goal with
  | |- OCit ?n _ => simple apply (k n)
  | _            => fail "Conclusion does not start with !^(_)"
  end.

Ltac lin_prom_it := (* usual promotion rule *)
  are_oc_hyps ;
  remocit_goal addocit_FORBIDDEN_IN_PROOFS ||
  (simpl ; remocit_goal addocit_FORBIDDEN_IN_PROOFS).

(*= Iterated promotion for ELL =*)

Fact ocit_fprom_FORBIDDEN_IN_PROOFS :
  (* Invalid promotion principle, to be used in an empty context only *)
     forall n : nat, forall a b : Prop, (a -> b) -> OCit n a -> OCit n b.
Proof.
intros n a b i.
induction n ; (* meta induction *)
simpl.
assumption.
apply oc_fprom_FORBIDDEN_IN_PROOFS.
assumption.
Qed.

Ltac lin_fprom_ocit :=
  (* iterated functorial promotion *)
  lazymatch goal with
  | H : OCit _ _ |- _ => revert H ;
                         no_hyps ;
                         (apply ocit_fprom_FORBIDDEN_IN_PROOFS ||
                         (simpl ; apply ocit_fprom_FORBIDDEN_IN_PROOFS )) ;
                         intro H
  end.

Fact ocit_mon_m : forall n : nat, forall a b : Prop, OCit n a -> OCit n b -> OCit n (a /\ b).
Proof.
induction n ;
simpl ;
intros a b Ha Hb.
lin_split with Ha ; lin_assumption.
rewrite ococit_ocitoc in Ha.
rewrite ococit_ocitoc in Hb.
rewrite ococit_ocitoc.
specialize IHn with (!a) (!b).
lin_hyp_apply2 IHn in Ha with Hb.
lin_fprom_ocit.
lin_left_conj Ha.
lem_apply2 oc_mon_m with Ha0 ;
lin_assumption.
lin_assumption.
Qed.

Fact ocit_mon_n : forall n : nat, True -> OCit n True.
Proof.
induction n ;
simpl ;
intro o.
lin_assumption.
lin_hyp_apply1 IHn in o.
rewrite ococit_ocitoc.
lin_fprom_ocit.
lem_apply1 oc_mon_n.
lin_assumption.
Qed.

Ltac lin_ufprom_ocit :=
  (* functorial promotion allowing empty exponential context *)
  lazymatch goal with
  | H : OCit _ _ |- _ => revert H ;
                         no_hyps ;
                         (apply ocit_fprom_FORBIDDEN_IN_PROOFS ||
                         (simpl ; apply ocit_fprom_FORBIDDEN_IN_PROOFS )) ;
                         intro H
  | |- OCit ?n _      => let H := fresh in
                         assert (OCit n True) as H ; [ lem_apply1 ocit_mon_n ; lin_one | idtac] ;
                         lin_ufprom_ocit ;
                         lin_left_one H
  end.

Ltac lin_mfprom_ocit :=
  (* multifunctorial promotion *)
  lazymatch goal with
  | H1 : OCit ?n _, H2 : OCit ?n _ |- _ => simple apply (ocit_mon_m _ _ _ H1) in H2 ; clear H1 ;
                                           lin_mfprom_ocit ;
                                           destruct H2 as [H1 H2]
  | H : OCit _ _ |- _                   => revert H ;
                                           no_hyps ;
                                           (apply ocit_fprom_FORBIDDEN_IN_PROOFS ||
                                           (simpl ; apply ocit_fprom_FORBIDDEN_IN_PROOFS )) ;
                                           intro H
  | |- OCit ?n _                        => let H := fresh in
                                           assert (OCit n True) as H ;
                                           [ lem_apply1 ocit_mon_n ; lin_one | idtac] ;
                                           lin_mfprom_ocit ;
                                           lin_left_one H
  end.

(*= Iterated promotion for LLL =*)

Fact pgit_fprom_FORBIDDEN_IN_PROOFS :
  (* Invalid promotion principle, to be used in an empty context only *)
     forall n : nat, forall a b : Prop, (a -> b) -> PGit n a -> PGit n b.
Proof.
intros n a b i.
induction n ; (* meta induction *)
simpl.
assumption.
apply pg_fprom_FORBIDDEN_IN_PROOFS.
assumption.
Qed.

Ltac lin_fprom_pgit :=
  (* iterated functorial promotion *)
  lazymatch goal with
  | H : PGit _ _ |- _ => revert H ;
                         no_hyps ;
                         (apply pgit_fprom_FORBIDDEN_IN_PROOFS ||
                         (simpl ; apply pgit_fprom_FORBIDDEN_IN_PROOFS )) ;
                         intro H
  end.

Fact pgit_mon_m : forall n : nat, forall a b : Prop, PGit n a -> PGit n b -> PGit n (a /\ b).
Proof.
induction n ;
simpl ;
intros a b Ha Hb.
lin_split with Ha ; lin_assumption.
rewrite pgpgit_pgitpg in Ha.
rewrite pgpgit_pgitpg in Hb.
rewrite pgpgit_pgitpg.
specialize IHn with (§a) (§b).
lin_hyp_apply2 IHn in Ha with Hb.
lin_fprom_pgit.
lin_left_conj Ha.
lem_apply2 pg_mon_m with Ha0 ;
lin_assumption.
lin_assumption.
Qed.

Fact pgit_mon_n : forall n : nat, True -> PGit n True.
Proof.
induction n ;
simpl ;
intro o.
lin_assumption.
lin_hyp_apply1 IHn in o.
rewrite pgpgit_pgitpg.
lin_fprom_pgit.
lem_apply1 pg_mon_n.
lin_assumption.
Qed.

Ltac lin_ufprom_pgit :=
  (* functorial promotion allowing empty exponential context *)
  lazymatch goal with
  | H : PGit _ _ |- _ => revert H ;
                         no_hyps ;
                         (apply pgit_fprom_FORBIDDEN_IN_PROOFS ||
                         (simpl ; apply pgit_fprom_FORBIDDEN_IN_PROOFS )) ;
                         intro H
  | |- PGit ?n _      => let H := fresh in
                         assert (PGit n True) as H ; [ lem_apply1 pgit_mon_n ; lin_one | idtac] ;
                         lin_ufprom_pgit ;
                         lin_left_one H
  end.

Ltac lin_mfprom_pgit :=
  (* multifunctorial promotion *)
  lazymatch goal with
  | H : OCit _ _ |- _                   => simple apply pgit_der in H ; lin_mfprom_pgit
  | H1 : PGit ?n _, H2 : PGit ?n _ |- _ => simple apply (pgit_mon_m _ _ _ H1) in H2 ; clear H1 ;
                                           lin_mfprom_pgit ;
                                           destruct H2 as [H1 H2]
  | H : PGit _ _ |- _                   => revert H ;
                                           no_hyps ;
                                           (apply pgit_fprom_FORBIDDEN_IN_PROOFS ||
                                           (simpl ; apply pgit_fprom_FORBIDDEN_IN_PROOFS )) ;
                                           intro H
  | |- PGit ?n _                        => let H := fresh in
                                           assert (PGit n True) as H ;
                                           [ lem_apply1 pgit_mon_n ; lin_one | idtac] ;
                                           lin_mfprom_pgit ;
                                           lin_left_one H
  end.

Ltac lin_der_pgit H := lem_hyp_apply1 pgit_der in H.
Ltac lin_right_der_pgit := lem_apply1 pgit_der.


(* ML modules for checking linearity before Qed. *)
(* llcheck
Declare ML Module "lincheck".
Declare ML Module "linqed".
llcheck *)
