(*== Encoding of Polytime Turing machines in Elementary Linear Logic ==*)
(*== with recursive types and data types containing garbage          ==*)

(* v.0.4 Olivier Laurent *)

Require Import stdlib_ell2coq. (* v.0.5.6-1 *)
Require Import ellchurch.      (* v.0.3 *)
Require Import rectypes.       (* v.0.2 *)


(*= Garbage =*)

Definition Garb := exists G : Prop, G.

Fact egarb : forall A : Prop, A -> Garb.
Proof.
intros A HA.
exists A.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact doubleegarb : forall A B : Prop, A -> B -> Garb.
Proof.
intros A B HA HB.
lem_apply1 (egarb (A /\ B)).
lin_split with HA ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.


(*= Booleans with and without garbage =*)

Definition LBool := (forall X : Prop, X -> X -> X /\ X).
Definition BoolG := (forall X : Prop, X -> X -> (X /\ Garb)).

Fact lbool2boolg : LBool -> BoolG.
intro b.
intros X x y.
specialize (b X).
lin_left_imp b with y.
lin_hyp_apply1 b in y.
lin_left_conj y.
lin_split with y0.
lin_assumption.
lem_apply1 (egarb X).
lin_assumption.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact ltrue : LBool.
Proof.
intros X x y.
lin_split with x ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact lfalse : LBool.
Proof.
intros X x y.
lin_split with y ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact trueG : BoolG.
Proof.
lem_apply1 lbool2boolg.
lem_apply0 ltrue.
(* llcheck
ELL.
llcheck *)
Qed.

Fact falseG : BoolG.
Proof.
lem_apply1 lbool2boolg.
lem_apply0 lfalse.
(* llcheck
ELL.
llcheck *)
Qed.

Fact extract_garb : BoolG -> LBool /\ Garb.
Proof.
intro b.
specialize (b LBool).
lin_apply2 b with.
lem_apply0 ltrue.
lem_apply0 lfalse.
(* llcheck
ELL.
llcheck *)
Qed.

Fact garbBool : forall G : Prop, BoolG /\ G -> BoolG.
Proof.
intros G (b,g) X x y.
lin_cut (X /\ Garb) as b' with g.
lin_left_conj b'.
lin_split with b'0.
lin_assumption.
lem_apply2 (doubleegarb (Garb) G) with b'1 ;
lin_assumption.
lin_apply2 b with x ;
lin_assumption.
(* llcheck
(* ELL. *)
(* BUG checker *)
llcheck *)
Qed.


(*= Four values set with garbage =*)

Definition QuadG := (forall X : Prop, X -> X -> X -> X -> (X /\ Garb)).

Fact one : QuadG.
Proof.
intros X x y z t.
lin_split with x.
lin_assumption.
lem_apply1 (egarb (X /\ X /\ X)).
lin_split with y.
lin_assumption.
lin_split with z ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact two : QuadG.
Proof.
intros X x y z t.
lin_split with y.
lin_assumption.
lem_apply1 (egarb (X /\ X /\ X)).
lin_split with x.
lin_assumption.
lin_split with z ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact three : QuadG.
Proof.
intros X x y z t.
lin_split with z.
lin_assumption.
lem_apply1 (egarb (X /\ X /\ X)).
lin_split with x.
lin_assumption.
lin_split with y ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact four : QuadG.
Proof.
intros X x y z t.
lin_split with t.
lin_assumption.
lem_apply1 (egarb (X /\ X /\ X)).
lin_split with x.
lin_assumption.
lin_split with y ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact is_four : QuadG -> BoolG.
Proof.
intro q.
lem_apply1 (garbBool Garb).
lin_apply4 q with :: ::.
lem_apply0 falseG.
lem_apply0 falseG.
lem_apply0 falseG.
lem_apply0 trueG.
(* llcheck
ELL.
llcheck *)
Qed.

Fact garbQuad : forall G : Prop, QuadG /\ G -> QuadG.
Proof.
intros G (q,g) X x y z t.
lin_cut (X /\ Garb) as q' with g.
lin_left_conj q'.
lin_split with q'0.
lin_assumption.
lem_apply2 (doubleegarb (Garb) G) with q'1 ;
lin_assumption.
lin_apply4 q with x :: y :: z ;
lin_assumption.
(* llcheck
(* ELL. *)
(* BUG checker *)
llcheck *)
Qed.


(*= Scott binary stacks with garbage =*)

Definition SBinG :=
  µ.(fun S => forall X : Prop, (S -> X) -> (S -> X) -> X -> (X /\ Garb)).
(* general stack type with elements in E:
  fun S => forall X : Prop, (E -> S -> X) -> X -> (X /\ Garb)
*)

Fact snil : SBinG.
Proof.
tunfold_goal.
intros X c0 c1 n.
lin_split with n.
lin_assumption.
lem_apply2 (doubleegarb (SBinG -> X) (SBinG -> X)) with c0 ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact scons0 : SBinG -> SBinG.
Proof.
intros s.
tunfold_goal.
intros X c0 c1 n.
lin_split with c0 s.
lin_apply1 c0.
lin_assumption.
lem_apply2 (doubleegarb (SBinG -> X) X) with c1 ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact scons1 : SBinG -> SBinG.
Proof.
intros s.
tunfold_goal.
intros X c0 c1 n.
lin_split with c1 s.
lin_apply1 c1.
lin_assumption.
lem_apply2 (doubleegarb (SBinG -> X) X) with c0 ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact pop : SBinG -> BoolG /\ SBinG.
Proof.
intro s.
lin_cut ((BoolG /\ SBinG) /\ Garb) as H with.
lin_left_conj H.
lin_left_conj H0.
lin_split with H2 H1.
lem_apply1 (garbBool Garb).
lin_split with H2 ;
lin_assumption.
lin_assumption.
hyp_tunfold s.
lin_apply3 s with ::.
intro t.
lin_split with.
lem_apply0 trueG.
lin_assumption.
intro t.
lin_split with.
lem_apply0 falseG.
lin_assumption.
lin_split with.
lem_apply0 trueG.
lem_apply0 snil.
(* llcheck
ELL.
llcheck *)
Qed.

Fact cbin2sbing : CBin -> !SBinG.
Proof.
intro l.
lin_cut (!SBinG) as i with l.
revert i.
iter2 l.
lem_apply0 scons0.
lem_apply0 scons1.
lin_mfprom_oc.
lem_apply0 snil.
(* llcheck
ELL.
llcheck *)
Qed.


(*= Polytime Turing machines =*)

(* Polynomial bounding the execution time of the Turing machine *)
Parameter poly : !CNat -> !CNat.

(* Examples of polynomials *)
Fact poly1 : !CNat -> !CNat. (* 2 X *)
Proof.
intro x.
cpoly (cons (2,1) nil).
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact poly2 : !CNat -> !CNat. (* 2 X^2 + 3 X + 1 *)
Proof.
intro x.
cpoly (cons (2,2) (cons (3,1) (cons (1,0) nil))).
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.


(* Definition of meta parameters in Set *)
Inductive IBool : Set := ITrue | IFalse.
  (* two symbols on tape *)
Inductive IQuad : Set := IOne | ITwo | IThree | IFour.
  (* four states for the machine *)
Inductive IMove : Set := ILeft | IRight.
  (* two possible moves *)


(* representation of the transition table of the Turing machine *)
Parameter TMstate : IQuad -> IBool -> IQuad.     (* next state *)
Parameter TMcell  : IQuad -> IBool -> IBool.     (* next symbol *)
Parameter TMmove  : IQuad -> IBool -> IMove.     (* move *)
(*
Fact TMstate : IQuad -> IBool -> IQuad.
Proof.
intros q a.
case q ; case a.
apply IThree.  (* 1 / 0 -> 3 *)
apply ITwo.    (* 1 / 1 -> 2 *)
apply IOne.    (* 2 / 0 -> 1 *)
apply IOne.    (* 2 / 1 -> 1 *)
apply IFour.   (* 3 / 0 -> 4 *)
apply IFour.   (* 3 / 1 -> 4 *)
apply IThree.  (* 4 / 0 -> 3 *)
apply IThree.  (* 4 / 1 -> 3 *)
Qed.
Fact TMcell : IQuad -> IBool -> IBool.
Proof.
intros q a.
case q ; case a.
apply ITrue.  (* 1 / 0 -> 0 *)
apply IFalse. (* 1 / 1 -> 1 *)
apply IFalse. (* 2 / 0 -> 1 *)
apply ITrue.  (* 2 / 1 -> 0 *)
apply ITrue.  (* 3 / 0 -> 0 *)
apply IFalse. (* 3 / 1 -> 1 *)
apply ITrue.  (* 4 / 0 -> 0 *)
apply IFalse. (* 4 / 1 -> 1 *)
Qed.
Fact TMmove : IQuad -> IBool -> IMove.
Proof.
intros q a.
case q.
apply IRight. (* 1 / _ -> R *)
apply IRight. (* 2 / _ -> R *)
apply ILeft.  (* 3 / _ -> L *)
apply ILeft.  (* 4 / _ -> L *)
Qed.
*)

Fact ibool2cons : IBool -> (SBinG -> SBinG).
Proof.
intro ib.
case ib.
lem_apply0 scons0.
lem_apply0 scons1.
(* llcheck
(* ELL. *)
llcheck *)
Qed.

Fact iquad2quadg : IQuad -> QuadG.
Proof.
intro iq.
case iq.
lem_apply0 one.
lem_apply0 two.
lem_apply0 three.
lem_apply0 four.
(* llcheck
(* ELL. *)
llcheck *)
Qed.

Definition Config := (((SBinG /\ BoolG) /\ SBinG) /\ QuadG).
  (* left part of the tape, current cell, right part of the tape, state *)

Fact garbConfig : forall G : Prop, Config /\ G -> Config.
Proof.
intros G (c,g).
lin_left_conj c.
lin_split with c0.
lin_assumption.
lem_apply1 (garbQuad G).
lin_split with c1 ;
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact init : SBinG -> Config.
Proof.
intro s.
lem_hyp_apply1 pop in s.
lin_split with s.
lin_left_conj s.
lin_split with s0.
lin_split with.
lem_apply0 snil.
lin_assumption.
lin_assumption.
lem_apply0 one.
(* llcheck
ELL.
llcheck *)
Qed.

Fact accept : Config -> LBool /\ Garb.
Proof.
intro c.
lem_apply1 extract_garb.
lin_left_conj c.
lem_apply1 (garbBool ((SBinG /\ BoolG) /\ SBinG)).
lin_split with c1.
lem_apply1 is_four.
lin_assumption.
lin_assumption.
(* llcheck
ELL.
llcheck *)
Qed.

Fact next_config : IQuad -> IBool -> SBinG -> SBinG -> Config.
  (* compute the next configuration from the left tape and the right tape 
     thanks to parameters in Set: current state and current symbol *)
Proof. 
(* preliminaries at the meta level in Set *)
intros istate icell.
assert (inext_state := TMstate istate icell).  (* meta computation in Set *)
assert (inext_cell  := TMcell  istate icell).  (* meta computation in Set *)
assert (imove       := TMmove  istate icell).  (* meta computation in Set *)
lin_clear istate.                              (* meta computation in Set *)
lin_clear icell.                               (* meta computation in Set *)
case imove ; lin_clear imove.                  (* meta computation in Set *)

(* move to left *)
intros lefttape righttape.
lin_split with lefttape righttape.
lin_split with lefttape.
lem_apply1 tens_comm.
lem_apply1 pop.
lin_assumption.
lem_apply1 (ibool2cons inext_cell).
lin_assumption.
lem_apply0 (iquad2quadg inext_state).

(* move to right *)
intros lefttape righttape.
lin_split with lefttape righttape.
lem_apply1 tens_assoc.
lin_split with lefttape.
lem_apply1 (ibool2cons inext_cell).
lin_assumption.
lem_apply1 pop.
lin_assumption.
lem_apply0 (iquad2quadg inext_state).
(* llcheck
(* ELL. *)
llcheck *)
Qed.

Fact transit : Config -> Config.
Proof.
intro c.
lin_left_conj c.
revert c0.
lin_cut (((SBinG /\ BoolG) /\ SBinG -> Config) /\ Garb) as g with.
lin_left_conj g.
intro tape.
lem_apply1 (garbConfig Garb).
lin_split with g0 tape.
lin_apply1 g0.
lin_assumption.
lin_assumption.
lin_apply4 c1 with :: :: ;
(
intros ((lefttape,cell),righttape) ;
revert lefttape righttape ;
lin_cut ((SBinG -> SBinG -> Config) /\ Garb) as g with ;
[
lin_left_conj g ;
intros lefttape righttape ;
lem_apply1 (garbConfig Garb) ;
lin_split with g0 lefttape righttape ;
  [ lin_apply2 g0 with lefttape ; lin_assumption
  | lin_assumption ]
|
lin_apply2 cell with
]
).
lem_apply0 (next_config IOne ITrue).
lem_apply0 (next_config IOne IFalse).
lem_apply0 (next_config ITwo ITrue).
lem_apply0 (next_config ITwo IFalse).
lem_apply0 (next_config IThree ITrue).
lem_apply0 (next_config IThree IFalse).
lem_apply0 (next_config IFour ITrue).
lem_apply0 (next_config IFour IFalse).
(* llcheck
ELL.
llcheck *)
Qed.

Fact run : !CBin -> !!(LBool /\ Garb).
Proof.
intro l.
lin_cut (!!Config) as r with.
lin_mfprom_oc.
lin_mfprom_oc.
lem_apply1 accept.
lin_assumption.
lin_contr l.
lin_cut (!CNat) as n with l.
lin_mfprom_oc.
lin_cut (!Config) as i with n.
revert i.
iter n.
lem_apply0 transit.
lem_hyp_apply1 cbin2sbing in l.
lin_mfprom_oc.
lem_apply1 init.
lin_assumption.
lem_apply1 poly.
lin_mfprom_oc.
lem_apply1 clength.
lin_assumption.
(*
Defined.
Eval compute in run.
*)
(* llcheck
ELL.
llcheck *)
Qed.


(*= Output in an affine context =*)

Definition ABool := (forall X : Prop, X -> X -> X).

Fact lboolg2lbool : (LBool /\ Garb) -> LBool. (* affine *)
Proof.
intros (b,g).
clear g. (* affine *)
lin_assumption.
(* llcheck
Fail ELL.
EAL.
llcheck *)
Qed.

Fact lbool2abool : LBool -> ABool. (* affine *)
Proof.
intro b.
intros X x y.
specialize (b X).
lin_cut (X /\ X) as p with.
lin_left_conj p.
clear p1. (* affine *)
lin_assumption.
lin_apply2 b with x ;
lin_assumption.
(* llcheck
Fail ELL.
EAL.
llcheck *)
Qed.



